
# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p'  ./heat/common/exception.py
HEAT_EXCEPTIONS = [
    "HeatException",
    "MissingCredentialError",
    "AuthorizationFailure",
    "NotAuthenticated",
    "Forbidden",
    "NotAuthorized",
    "Invalid",
    "UserParameterMissing",
    "UnknownUserParameter",
    "InvalidTemplateVersion",
    "InvalidTemplateSection",
    "ImmutableParameterModified",
    "InvalidMergeStrategyForParam",
    "ConflictingMergeStrategyForParam",
    "InvalidTemplateAttribute",
    "InvalidTemplateReference",
    "TemplateOutputError",
    "InvalidEncryptionKey",
    "InvalidExternalResourceDependency",
    "EntityNotFound",
    "PhysicalResourceExists",
    "PhysicalResourceNameAmbiguity",
    "PhysicalResourceIDAmbiguity",
    "InvalidTenant",
    "StackExists",
    "HeatExceptionWithPath",
    "StackValidationFailed",
    "InvalidSchemaError",
    "ResourceNotFound",
    "SnapshotNotFound",
    "InvalidGlobalResource",
    "ResourceTypeUnavailable",
    "InvalidBreakPointHook",
    "InvalidRestrictedAction",
    "ResourceNotAvailable",
    "ClientNotAvailable",
    "ResourceFailure",
    "NotSupported",
    "ResourceActionNotSupported",
    "ResourceActionRestricted",
    "ResourcePropertyConflict",
    "ResourcePropertyDependency",
    "ResourcePropertyValueDependency",
    "PropertyUnspecifiedError",
    "UpdateReplace",
    "ResourceUnknownStatus",
    "ResourceInError",
    "UpdateInProgress",
    "HTTPExceptionDisguise",
    "EgressRuleNotAllowed",
    "Error",
    "NotFound",
    "InvalidContentType",
    "RequestLimitExceeded",
    "DownloadLimitExceeded",
    "StackResourceLimitExceeded",
    "ActionInProgress",
    "ActionNotComplete",
    "StopActionFailed",
    "EventSendFailed",
    "InterfaceAttachFailed",
    "InterfaceDetachFailed",
    "UnsupportedObjectError",
    "OrphanedObjectError",
    "IncompatibleObjectVersion",
    "ObjectActionError",
    "ReadOnlyFieldError",
    "ConcurrentTransaction",
    "ObjectFieldInvalid",
    "KeystoneServiceNameConflict",
    "SIGHUPInterrupt",
    "InvalidServiceVersion",
    "InvalidTemplateVersions",
    "UnableToAutoAllocateNetwork",
]
