# flake8: noqa

# From https://github.com/openstack/manila/blob/master/manila/exception.py
# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' manila/exception.py
MANILA_EXCEPTIONS = [
    "ConvertedException",
    "Error",
    "ManilaException",
    "NetworkException",
    "NetworkBindException",
    "NetworkBadConfigurationException",
    "BadConfigurationException",
    "NotAuthorized",
    "AdminRequired",
    "PolicyNotAuthorized",
    "Conflict",
    "Invalid",
    "InvalidRequest",
    "InvalidResults",
    "InvalidInput",
    "InvalidContentType",
    "InvalidHost",
    "InvalidParameterValue",
    "InvalidUUID",
    "InvalidDriverMode",
    "InvalidAPIVersionString",
    "VersionNotFoundForAPIMethod",
    "InvalidGlobalAPIVersion",
    "InvalidCapacity",
    "NotFound",
    "MessageNotFound",
    "Found",
    "InUse",
    "AvailabilityZoneNotFound",
    "ShareNetworkNotFound",
    "ShareNetworkSubnetNotFound",
    "ShareServerNotFound",
    "ShareServerNotFoundByFilters",
    "InvalidShareNetwork",
    "ShareServerInUse",
    "ShareServerMigrationError",
    "ShareServerMigrationFailed",
    "InvalidShareServer",
    "ShareMigrationError",
    "ShareMigrationFailed",
    "ShareDataCopyFailed",
    "ShareDataCopyCancelled",
    "ServiceIPNotFound",
    "AdminIPNotFound",
    "ShareServerNotCreated",
    "ShareServerNotReady",
    "ServiceNotFound",
    "ServiceIsDown",
    "HostNotFound",
    "SchedulerHostFilterNotFound",
    "SchedulerHostWeigherNotFound",
    "HostBinaryNotFound",
    "InvalidReservationExpiration",
    "InvalidQuotaValue",
    "QuotaNotFound",
    "QuotaExists",
    "QuotaResourceUnknown",
    "ProjectUserQuotaNotFound",
    "ProjectShareTypeQuotaNotFound",
    "ProjectQuotaNotFound",
    "QuotaClassNotFound",
    "QuotaUsageNotFound",
    "ReservationNotFound",
    "OverQuota",
    "MigrationNotFound",
    "MigrationNotFoundByStatus",
    "FileNotFound",
    "MigrationError",
    "MalformedRequestBody",
    "ConfigNotFound",
    "PasteAppNotFound",
    "NoValidHost",
    "WillNotSchedule",
    "QuotaError",
    "ShareSizeExceedsAvailableQuota",
    "SnapshotSizeExceedsAvailableQuota",
    "ShareSizeExceedsLimit",
    "ShareLimitExceeded",
    "SnapshotLimitExceeded",
    "ShareNetworksLimitExceeded",
    "ShareGroupsLimitExceeded",
    "ShareGroupSnapshotsLimitExceeded",
    "ShareReplicasLimitExceeded",
    "ShareReplicaSizeExceedsAvailableQuota",
#    "GlusterfsException",
    "InvalidShare",
    "ShareBusyException",
    "InvalidShareInstance",
    "ManageInvalidShare",
    "ManageShareServerError",
    "UnmanageInvalidShare",
    "PortLimitExceeded",
    "ShareAccessExists",
    "ShareAccessMetadataNotFound",
    "ShareSnapshotAccessExists",
    "InvalidSnapshotAccess",
    "InvalidShareAccess",
    "InvalidShareAccessLevel",
    "InvalidShareAccessType",
    "ShareBackendException",
    "ExportLocationNotFound",
    "ShareNotFound",
    "ShareSnapshotNotFound",
    "ShareSnapshotInstanceNotFound",
    "ShareSnapshotNotSupported",
    "ShareGroupSnapshotNotSupported",
    "ShareSnapshotIsBusy",
    "InvalidShareSnapshot",
    "InvalidShareSnapshotInstance",
    "ManageInvalidShareSnapshot",
    "UnmanageInvalidShareSnapshot",
    "ShareMetadataNotFound",
    "InvalidMetadata",
    "InvalidMetadataSize",
    "SecurityServiceNotFound",
    "InvalidSecurityService",
    "ShareNetworkSecurityServiceAssociationError",
    "ShareNetworkSecurityServiceDissociationError",
    "SecurityServiceFailedAuth",
    "InvalidVolume",
    "InvalidShareType",
    "InvalidShareGroupType",
    "InvalidExtraSpec",
    "VolumeNotFound",
    "VolumeSnapshotNotFound",
    "ShareTypeNotFound",
    "ShareGroupTypeNotFound",
    "ShareTypeAccessNotFound",
    "ShareGroupTypeAccessNotFound",
    "ShareTypeNotFoundByName",
    "ShareGroupTypeNotFoundByName",
    "ShareTypeExtraSpecsNotFound",
    "ShareGroupTypeSpecsNotFound",
    "ShareTypeInUse",
    "IPAddressInUse",
    "ShareGroupTypeInUse",
    "ShareTypeExists",
    "ShareTypeDoesNotExist",
    "DefaultShareTypeNotConfigured",
    "ShareGroupTypeExists",
    "ShareTypeAccessExists",
    "ShareGroupTypeAccessExists",
    "ShareTypeCreateFailed",
    "ShareTypeUpdateFailed",
    "ShareGroupTypeCreateFailed",
    "ManageExistingShareTypeMismatch",
    "ShareExtendingError",
    "ShareShrinkingError",
    "ShareShrinkingPossibleDataLoss",
    "InstanceNotFound",
    "BridgeDoesNotExist",
    "ServiceInstanceException",
    "ServiceInstanceUnavailable",
    "StorageResourceException",
    "StorageResourceNotFound",
    "SnapshotResourceNotFound",
    "SnapshotUnavailable",
#    "NetAppException",
#    "VserverNotFound",
#    "VserverNotSpecified",
#    "VserverNotReady",
#    "EMCPowerMaxXMLAPIError",
#    "EMCPowerMaxLockRequiredException",
#    "EMCPowerMaxInvalidMoverID",
#    "EMCVnxXMLAPIError",
#    "EMCVnxLockRequiredException",
#    "EMCVnxInvalidMoverID",
#    "EMCUnityError",
#    "HPE3ParInvalidClient",
#    "HPE3ParInvalid",
#    "HPE3ParUnexpectedError",
#    "GPFSException",
#    "GPFSGaneshaException",
    "GaneshaCommandFailure",
    "InvalidSqliteDB",
    "SSHException",
#    "HDFSException",
#    "MapRFSException",
#    "ZFSonLinuxException",
    "QBException",
    "QBRpcException",
    "SSHInjectionThreat",
#    "HNASBackendException",
#    "HNASConnException",
#    "HNASSSCIsBusy",
#    "HNASSSCContextChange",
#    "HNASDirectoryNotEmpty",
#    "HNASItemNotFoundException",
#    "HNASNothingToCloneException",
    "ShareGroupNotFound",
    "ShareGroupSnapshotNotFound",
    "ShareGroupSnapshotMemberNotFound",
    "InvalidShareGroup",
    "InvalidShareGroupSnapshot",
    "DriverNotInitialized",
    "ShareResourceNotFound",
    "ShareUmountException",
    "ShareMountException",
    "ShareCopyDataException",
    "ReplicationException",
    "ShareReplicaNotFound",
#    "TegileAPIException",
    "StorageCommunicationException",
    "EvaluatorParseException",
#    "HSPBackendException",
#    "HSPTimeoutException",
#    "HSPItemNotFoundException",
#    "NexentaException",
    "LockCreationFailed",
    "LockingFailed",
    "GaneshaException",
#    "InfortrendCLIException",
#    "InfortrendNASException",
#    "ZadaraUnknownCmd",
#    "ZadaraSessionRequestException",
#    "ZadaraBadHTTPResponseStatus",
#    "ZadaraFailedCmdWithDump",
#    "ZadaraVPSANoActiveController",
#    "ZadaraServerCreateFailure",
#    "ZadaraAttachmentsNotFound",
#    "ZadaraManilaInvalidAccessKey",
#    "ZadaraVPSAVolumeShareFailed",
#    "ZadaraInvalidShareAccessType",
#    "ZadaraShareNotFound",
#    "ZadaraExtendShareFailed",
#    "ZadaraInvalidProtocol",
#    "ZadaraShareNotValid",
#    "ZadaraVPSASnapshotCreateFailed",
#    "ZadaraVPSASnapshotManageFailed",
#    "ZadaraServerNotFound",
]
